@@ -5,6 +5,7 @@ import os
 import re
 import subprocess
 from io import open
+from pathlib import Path
 from pprint import pformat as pretty
 from typing import Any, Dict, Optional, Protocol
 
@@ -35,7 +36,7 @@ class AdjustCallback(Protocol):
     """
     A callback for per-rule notifications made by Tree.adjust()
 
-    Function be be called for every rule inspected by adjust().
+    Function which will be called for every rule inspected by adjust().
     It will be given three arguments: fmf tree being inspected,
     current adjust rule, and whether the rule was skipped (``None``),
     applied (``True``) or not applied (``False``).
@@ -49,6 +50,22 @@ class AdjustCallback(Protocol):
         pass
 
 
+class ApplyRulesCallback(Protocol):
+    """
+    A callback to decide if rules should be processed in Tree.adjust()
+
+    Function to be called for every node before ``additional_rules``
+    are processed. It is called with fmf tree as the parameter.
+    It should return ``True`` when additional_rules should be processed
+    or ``False`` when they should be ignored.
+    """
+
+    def __call__(
+            self,
+            node: 'Tree') -> bool:
+        return True
+
+
 class Tree:
     """ Metadata Tree """
 
@@ -71,6 +88,7 @@ class Tree:
         self.data = dict()
         self.sources = list()
         self.root = None
+        self.config = {}
         self.version = utils.VERSION
         self.original_data = dict()
         self._commit = None
@@ -98,6 +116,7 @@ class Tree:
         # Handle child node creation
         else:
             self.root = self.parent.root
+            self.config = self.parent.config
             self.name = os.path.join(self.parent.name, name)
 
         # Update data from a dictionary (handle empty nodes)
@@ -145,7 +164,8 @@ class Tree:
         return self.name
 
     def _initialize(self, path):
-        """ Find metadata tree root, detect format version """
+        """ Find metadata tree root, detect format version, check for config """
+
         # Find the tree root
         root = os.path.abspath(path)
         try:
@@ -159,6 +179,7 @@ class Tree:
             raise utils.FileError("Invalid directory path: {0}".format(root))
         log.info("Root directory found: {0}".format(root))
         self.root = root
+
         # Detect format version
         try:
             with open(os.path.join(self.root, ".fmf", "version")) as version:
@@ -170,16 +191,55 @@ class Tree:
         except ValueError:
             raise utils.FormatError("Invalid version format")
 
+        # Check for the config file
+        config_file_path = Path(self.root) / ".fmf/config"
+        try:
+            self.config = YAML(typ="safe").load(config_file_path.read_text())
+            log.debug(f"Config file '{config_file_path}' loaded.")
+        except FileNotFoundError:
+            log.debug("Config file not found.")
+        except YAMLError as error:
+            raise utils.FileError(f"Failed to parse '{config_file_path}'.\n{error}")
+
     def _merge_plus(self, data, key, value, prepend=False):
         """ Handle extending attributes using the '+' suffix """
-        # Nothing to do if key not in parent
+
+        # Set the value if key is not present
         if key not in data:
             data[key] = value
             return
+
+        # Parent is a list of dict and child is a dict
+        # (just update every parent list item using the special merge)
+        if isinstance(data[key], list) and isinstance(value, dict):
+            for list_item in data[key]:
+                if not isinstance(list_item, dict):
+                    raise utils.MergeError(
+                        "MergeError: Item '{0}' in {1} must be a dictionary.".format(
+                            list_item, self.name))
+                self._merge_special(list_item, value)
+            return
+
+        # Parent is a dict and child is a list of dict
+        # (replace parent dict with the list of its special-merged copies)
+        if isinstance(data[key], dict) and isinstance(value, list):
+            result_list = []
+            for list_item in value:
+                if not isinstance(list_item, dict):
+                    raise utils.MergeError(
+                        "MergeError: Item '{0}' in {1} must be a dictionary.".format(
+                            list_item, self.name))
+                result_dict = copy.deepcopy(data[key])
+                self._merge_special(result_dict, list_item)
+                result_list.append(result_dict)
+            data[key] = result_list
+            return
+
         # Use the special merge for merging dictionaries
-        if type(data[key]) == type(value) == dict:
+        if isinstance(data[key], dict) and isinstance(value, dict):
             self._merge_special(data[key], value)
             return
+
         # Attempt to apply the plus operator
         try:
             if prepend:
@@ -193,6 +253,9 @@ class Tree:
 
     def _merge_regexp(self, data, key, value):
         """ Handle substitution of current values """
+        # Nothing to substitute if the key is not present in parent
+        if key not in data:
+            return
         if isinstance(value, str):
             value = [value]
         for pattern, replacement in [utils.split_pattern_replacement(v) for v in value]:
@@ -218,6 +281,9 @@ class Tree:
                 if re.search(p, str(item)):
                     return True
             return False
+        # Nothing to remove if the key is not present in parent
+        if key not in data:
+            return
         if isinstance(value, str):
             value = [value]
         if isinstance(data[key], list):
@@ -238,18 +304,15 @@ class Tree:
         """ Handle reducing attributes using the '-' suffix """
         # Cannot reduce attribute if key is not present in parent
         if key not in data:
-            data[key] = value
-            raise utils.MergeError(
-                "MergeError: Key '{0}' in {1} (not inherited).".format(
-                    key, self.name))
+            return
         # Subtract numbers
-        if type(data[key]) == type(value) in [int, float]:
+        if isinstance(data[key], (int, float)) and isinstance(value, (int, float)):
             data[key] = data[key] - value
         # Replace matching regular expression with empty string
-        elif type(data[key]) == type(value) == type(""):
+        elif isinstance(data[key], str) and isinstance(value, str):
             data[key] = re.sub(value, '', data[key])
         # Remove given values from the parent list
-        elif type(data[key]) == type(value) == list:
+        elif isinstance(data[key], list) and isinstance(value, list):
             data[key] = [item for item in data[key] if item not in value]
         # Remove given key from the parent dictionary
         elif isinstance(data[key], dict) and isinstance(value, list):
@@ -262,7 +325,7 @@ class Tree:
 
     def _merge_special(self, data, source):
         """ Merge source dict into data, handle special suffixes """
-        for key, value in sorted(source.items()):
+        for key, value in source.items():
             # Handle special attribute merging
             if key.endswith('+'):
                 self._merge_plus(data, key.rstrip('+'), value)
@@ -325,17 +388,15 @@ class Tree:
 
     def merge(self, parent=None):
         """ Merge parent data """
-        # Check parent, append source files
+        # Check parent
         if parent is None:
             parent = self.parent
-        if parent is None:
-            return
-        # Do not inherit when disabled
-        if self._directives.get("inherit") is False:
-            return
-        self.sources = parent.sources + self.sources
-        # Merge child data with parent data
-        data = copy.deepcopy(parent.data)
+        if self._directives.get("inherit") is False or parent is None:
+            # Nothing to inherit
+            data = {}
+        else:
+            self.sources = parent.sources + self.sources
+            data = copy.deepcopy(parent.data)
         self._merge_special(data, self.data)
         self.data = data
 
@@ -368,7 +429,7 @@ class Tree:
             pass
 
         # Process the metadata
-        for key, value in sorted(data.items()):
+        for key, value in data.items():
             # Ensure there are no 'None' keys
             if key is None:
                 raise utils.FormatError("Invalid key 'None'.")
@@ -389,9 +450,15 @@ class Tree:
         log.debug("Data for '{0}' updated.".format(self))
         log.data(pretty(self.data))
 
-    def adjust(self, context, key='adjust', undecided='skip',
-               case_sensitive=True, decision_callback=None,
-               additional_rules=None):
+    def adjust(
+            self,
+            context,
+            key='adjust',
+            undecided='skip',
+            case_sensitive=True,
+            decision_callback: Optional[AdjustCallback] = None,
+            additional_rules=None,
+            additional_rules_callback: Optional[ApplyRulesCallback] = None):
         """
         Adjust tree data based on provided context and rules
 
@@ -418,6 +485,12 @@ class Tree:
         that should be applied after those from the node itself.
         These additional rules are processed even when an applied
         rule defined in the node has ``continue: false`` set.
+
+        Optional 'additional_rules_callback' callback could be set to
+        limit nodes for which 'additional_rules' are processed. This
+        callback is called with the current fmf node as an argument and
+        should return 'True' to process 'additional_rules' or 'False' to
+        skip them.
         """
 
         # Check context sanity
@@ -448,9 +521,8 @@ class Tree:
 
         context.case_sensitive = case_sensitive
 
-        # 'continue' has to affect only its rule_set
-        for rule_set in rules, additional_rules:
-            # Check and apply each rule
+        def apply_rules(rule_set):
+            # 'continue' has to affect only its rule_set
             for rule in rule_set:
                 # Rule must be a dictionary
                 if not isinstance(rule, dict):
@@ -503,12 +575,19 @@ class Tree:
                             "Invalid value for the 'undecided' parameter. Should "
                             "be 'skip' or 'raise', got '{}'.".format(undecided))
 
+        # Always process rules from 'key' (adjust)
+        apply_rules(rules)
+        # Additional rules might be skipped depending on the callback
+        if additional_rules_callback is None or additional_rules_callback(self):
+            apply_rules(additional_rules)
+
         # Adjust all child nodes as well
         for child in self.children.values():
             child.adjust(context, key, undecided,
                          case_sensitive=case_sensitive,
                          decision_callback=decision_callback,
-                         additional_rules=additional_rules)
+                         additional_rules=additional_rules,
+                         additional_rules_callback=additional_rules_callback)
 
     def get(self, name=None, default=None):
         """
@@ -555,6 +634,21 @@ class Tree:
             self.children[name].sources.append(source)
             self.children[name]._raw_data = copy.deepcopy(data)
 
+    @property
+    def explore_include(self):
+        """ Additional filenames to be explored """
+        try:
+            explore_include = self.config["explore"]["include"]
+            if not isinstance(explore_include, list):
+                raise utils.GeneralError(
+                    f"The 'include' config section should be a list, found '{explore_include}'.")
+            if ".fmf" in explore_include:
+                raise utils.GeneralError(
+                    "The '.fmf' directory cannot be used for storing fmf metadata.")
+        except KeyError:
+            explore_include = []
+        return explore_include
+
     def grow(self, path):
         """
         Grow the metadata tree for the given directory path
@@ -575,6 +669,7 @@ class Tree:
         except StopIteration:
             log.debug("Skipping '{0}' (not accessible).".format(path))
             return
+
         # Investigate main.fmf as the first file (for correct inheritance)
         filenames = sorted(
             [filename for filename in filenames if filename.endswith(SUFFIX)])
@@ -582,9 +677,10 @@ class Tree:
             filenames.insert(0, filenames.pop(filenames.index(MAIN)))
         except ValueError:
             pass
+
         # Check every metadata file and load data (ignore hidden)
         for filename in filenames:
-            if filename.startswith("."):
+            if filename.startswith(".") and filename not in self.explore_include:
                 continue
             fullpath = os.path.abspath(os.path.join(dirpath, filename))
             log.info("Checking file {0}".format(fullpath))
@@ -605,9 +701,10 @@ class Tree:
             # Handle other *.fmf files as children
             else:
                 self.child(os.path.splitext(filename)[0], data, fullpath)
+
         # Explore every child directory (ignore hidden dirs and subtrees)
         for dirname in sorted(dirnames):
-            if dirname.startswith("."):
+            if dirname.startswith(".") and dirname not in self.explore_include:
                 continue
             fulldir = os.path.join(dirpath, dirname)
             if os.path.islink(fulldir):
@@ -627,6 +724,7 @@ class Tree:
                 log.debug("Ignoring metadata tree '{0}'.".format(dirname))
                 continue
             self.child(dirname, os.path.join(path, dirname))
+
         # Ignore directories with no metadata (remove all child nodes which
         # do not have children and their data haven't been updated)
         for name in list(self.children.keys()):
@@ -635,12 +733,33 @@ class Tree:
                 del self.children[name]
                 log.debug("Empty tree '{0}' removed.".format(child.name))
 
-    def climb(self, whole=False):
-        """ Climb through the tree (iterate leaf/all nodes) """
+    def climb(self, whole: bool = False, sort: bool = True):
+        """
+        Climb through the tree (iterate over nodes)
+
+        :param whole: By default only leaf nodes are considered. When
+            set to ``True`` all nodes are iterated, including parent
+            branches.
+
+        :param sort: When iterating, child nodes are sorted by name by
+            default. Set to ``False`` if you prefer to keep the order in
+            which the child nodes were inserted into the tree.
+        """
+
+        # Include branches when `whole` is enabled or the `select`
+        # directive has been used to pick this node.
         if whole or self.select:
             yield self
-        for name, child in self.children.items():
-            for node in child.climb(whole):
+
+        # Sort child nodes by name only if requested
+        if sort:
+            children = [child for _, child in sorted(self.children.items())]
+        else:
+            children = self.children.values()
+
+        # Iterate through each child node
+        for child in children:
+            for node in child.climb(whole=whole, sort=sort):
                 yield node
 
     @property
@@ -658,9 +777,36 @@ class Tree:
                 return node
         return None
 
-    def prune(self, whole=False, keys=None, names=None, filters=None,
-              conditions=None, sources=None):
-        """ Filter tree nodes based on given criteria """
+    def prune(
+            self,
+            whole: bool = False,
+            keys: Optional[list[str]] = None,
+            names: Optional[list[str]] = None,
+            filters: Optional[list[str]] = None,
+            conditions: Optional[list[str]] = None,
+            sources: Optional[list[str]] = None,
+            sort: bool = True):
+        """
+        Filter tree nodes based on given criteria
+
+        :param whole: By default only leaf nodes are considered. When
+            set to ``True`` all nodes are iterated, including parent
+            branches.
+
+        :param keys: Include only nodes containing given keys.
+
+        :param names: Include only nodes matching provided names.
+
+        :param filters: Include only nodes matching given filters.
+
+        :param conditions: Include only nodes satisfying the conditions.
+
+        :param sources: Filter by source fmf file names on disk.
+
+        :param sort: When iterating, child nodes are sorted by name by
+            default. Set to ``False`` if you prefer to keep the order in
+            which the child nodes were inserted into the tree.
+        """
         keys = keys or []
         names = names or []
         filters = filters or []
@@ -670,7 +816,7 @@ class Tree:
         if sources:
             sources = {os.path.abspath(src) for src in sources}
 
-        for node in self.climb(whole):
+        for node in self.climb(whole, sort=sort):
             # Select only nodes with key content
             if not all([key in node.data for key in keys]):
                 continue
@@ -718,10 +864,10 @@ class Tree:
         # List available attributes
         for key, value in sorted(self.data.items()):
             output += "\n{0}: ".format(utils.color(key, 'green'))
-            if isinstance(value, type("")):
+            if isinstance(value, str):
                 output += value.rstrip("\n")
             elif isinstance(value, list) and all(
-                    [isinstance(item, type("")) for item in value]):
+                    [isinstance(item, str) for item in value]):
                 output += utils.listed(value)
             else:
                 output += pretty(value)
