@@ -34,15 +34,18 @@ class InvalidContext(Exception):
 class ContextValue:
     """ Value for dimension """
 
-    def __init__(self, origin):
+    def __init__(self, raw):
         """
         ContextValue("foo-1.2.3")
         ContextValue(["foo", "1", "2", "3"])
         """
-        if isinstance(origin, (tuple, list)):
-            self._to_compare = tuple(origin)
+        if isinstance(raw, (tuple, list)):
+            self._to_compare = tuple(raw)
         else:
-            self._to_compare = self._split_to_version(origin)
+            self._to_compare = self._split_to_version(raw)
+
+        # Store the original string for regexp processing
+        self.raw = raw
 
     def __eq__(self, other):
         if isinstance(other, self.__class__):
@@ -238,6 +241,22 @@ class Context:
 
         return self._op_core(dimension_name, values, comparator)
 
+    def _op_match(self, dimension_name, values):
+        """ '~' operator, regular expression matches """
+
+        def comparator(dimension_value, it_val):
+            return re.search(it_val.raw, dimension_value.raw) is not None
+
+        return self._op_core(dimension_name, values, comparator)
+
+    def _op_not_match(self, dimension_name, values):
+        """ '~' operator, regular expression does not match """
+
+        def comparator(dimension_value, it_val):
+            return re.search(it_val.raw, dimension_value.raw) is None
+
+        return self._op_core(dimension_name, values, comparator)
+
     def _op_minor_eq(self, dimension_name, values):
         """ '~=' operator """
 
@@ -371,20 +390,22 @@ class Context:
         "~>=": _op_minor_greater_or_equal,
         ">": _op_greater,
         "~>": _op_minor_greater,
+        "~": _op_match,
+        "!~": _op_not_match,
         }
 
     # Triple expression: dimension operator values
     # [^=].* is necessary as .+ matches '= something'
     re_expression_triple = re.compile(
-        r"(\w+)"
+        r"([\w-]+)"
         + r"\s*("
         + r"|".join(
-            set(operator_map.keys()) - {"is defined", "is not defined"})
+            [key for key in operator_map if key not in ["is defined", "is not defined"]])
         + r")\s*"
         + r"([^=].*)")
     # Double expression: dimension operator
     re_expression_double = re.compile(
-        r"(\w+)" + r"\s*(" + r"|".join(["is defined", "is not defined"]) + r")"
+        r"([\w-]+)" + r"\s*(" + r"|".join(["is defined", "is not defined"]) + r")"
         )
 
     # Simple boolean value
