@@ -4,6 +4,7 @@ import tempfile
 import threading
 import time
 from shutil import rmtree
+from textwrap import dedent
 
 import pytest
 from ruamel.yaml import YAML
@@ -48,6 +49,9 @@ class TestTree:
     def test_hidden(self):
         """ Hidden files and directories """
         assert ".hidden" not in self.wget.children
+        hidden = Tree(EXAMPLES + "hidden")
+        plan = hidden.find("/.plans/basic")
+        assert plan.get("discover") == {"how": "fmf"}
 
     def test_inheritance(self):
         """ Inheritance and data types """
@@ -77,6 +81,27 @@ class TestTree:
         child = Tree(EXAMPLES + "child")
         assert child.find("/nobody") is None
 
+    def test_insert_child(self):
+        """ Manual child creation """
+
+        # Prepare a simple tree by manually inserting child nodes
+        tree = Tree(data={"key": "value"})
+        tree.child(name="child2", data={"key": "value"})
+        tree.child(name="child3", data={"key": "value"})
+        tree.child(name="child1", data={"key": "value"})
+
+        # By default, node names should be sorted when climbing & prunning
+        expected = ['/child1', '/child2', '/child3']
+        assert [node.name for node in tree.climb()] == expected
+        assert [node.name for node in tree.prune()] == expected
+        assert [node.name for node in tree.climb(sort=True)] == expected
+        assert [node.name for node in tree.prune(sort=True)] == expected
+
+        # Original order should be kept if requested
+        expected = ['/child2', '/child3', '/child1']
+        assert [node.name for node in tree.climb(sort=False)] == expected
+        assert [node.name for node in tree.prune(sort=False)] == expected
+
     def test_prune_sources(self):
         """ Pruning by sources """
         original_directory = os.getcwd()
@@ -134,10 +159,33 @@ class TestTree:
         assert 'because' not in child.data
         assert 'foo' in child.data
 
+    def test_adjust_can_extend(self):
+        """ It is possible to extend dictionary with adjust """
+        tree = Tree(YAML(typ='safe').load(dedent("""
+        /a:
+         environment+:
+          FOO: bar
+         adjust+:
+          environment+:
+            BAR: baz
+        """)))
+        tree.adjust(context=Context())
+        child = tree.find('/a')
+        assert child.data['environment']['FOO'] == "bar"
+        assert child.data['environment']['BAR'] == "baz"
+
+    def test_node_without_parent_strips_merge_suffix(self):
+        """ Merge suffix is stripped in the top node as well """
+        tree = Tree(EXAMPLES + 'merge')
+        stray_child = tree.find('/stray')
+        assert 'environment' in stray_child.data
+        child = tree.find('/parent/reduced')
+        assert 'environment' in child.data
+
     def test_deep_hierarchy(self):
         """ Deep hierarchy on one line """
         deep = Tree(EXAMPLES + "deep")
-        assert len(deep.children) == 1
+        assert len(deep.children) == 3
 
     def test_deep_dictionary(self):
         """ Get value from a deep dictionary """
@@ -147,6 +195,17 @@ class TestTree:
         assert deep.get(['hardware', 'bad', 'size'], 12) == 12
         assert deep.get('nonexistent', default=3) == 3
 
+    def test_deep_dictionary_undefined_keys(self):
+        """ Extending undefined keys using '+' should work """
+        deep = Tree(EXAMPLES + "deep")
+        single = deep.find("/single")
+        assert single.get(["undefined", "deeper+", "key"]) == "value"
+        child = deep.find("/parent/child")
+        assert child.get("one") == 2
+        assert child.get("two") == 4
+        assert child.get("three") == 3
+        assert child.get(["undefined", "deeper+", "key"]) == "value"
+
     def test_merge_plus(self):
         """ Extending attributes using the '+' suffix """
         child = self.merge.find('/parent/extended')
@@ -161,6 +220,32 @@ class TestTree:
             child.data["time+"] = "string"
             child.inherit()
 
+    def test_merge_plus_parent_dict(self):
+        """ Merging parent dict with child list """
+        child = self.merge.find('/parent-dict/path')
+        assert len(child.data['discover']) == 2
+        assert child.data['discover'][0]['how'] == 'fmf'
+        assert child.data['discover'][0]['name'] == 'upstream'
+        assert child.data['discover'][0]['url'] == 'https://some.url'
+        assert child.data['discover'][0]['summary'] == 'test.upstream'
+        assert child.data['discover'][1]['how'] == 'fmf'
+        assert child.data['discover'][1]['name'] == 'downstream'
+        assert child.data['discover'][1]['url'] == 'https://other.url'
+        assert child.data['discover'][1]['summary'] == 'test.downstream'
+
+    def test_merge_plus_parent_list(self):
+        """ Merging parent list with child dict """
+        for i in [1, 2]:
+            child = self.merge.find(f'/parent-list/tier{i}')
+            assert child.data['summary'] == 'basic tests' if i == 1 else 'detailed tests'
+            assert len(child.data['discover']) == 2
+            assert child.data['discover'][0]['filter'] == f'tier: {i}'
+            assert child.data['discover'][0]['url'] == 'https://github.com/project1'
+            assert child.data['discover'][0]['summary'] == f'project1.tier{i}'
+            assert child.data['discover'][1]['filter'] == f'tier: {i}'
+            assert child.data['discover'][1]['url'] == 'https://github.com/project2'
+            assert child.data['discover'][1]['summary'] == f'project2.tier{i}'
+
     def test_merge_minus(self):
         """ Reducing attributes using the '-' suffix """
         child = self.merge.find('/parent/reduced')
@@ -170,10 +255,10 @@ class TestTree:
         assert child.data['time'] == 5
         assert child.data['vars'] == dict(x=1)
         assert 'time+' not in child.data
-        with pytest.raises(utils.MergeError):
-            child.data["disabled-"] = True
-            child.inherit()
-        child.data.pop('disabled-')
+        # Do not raise MergeError if key is missing
+        child.data["pkgs-"] = 'foo'
+        child.inherit()
+        assert 'pkgs-' not in child.data
         with pytest.raises(utils.MergeError):
             child.data["time-"] = "bad"
             child.inherit()
@@ -198,6 +283,13 @@ class TestTree:
         child = self.merge.find('/parent/buried')
         assert child.data['very']['deep']['dict'] == dict(x=2, y=1, z=0)
 
+    def test_merge_order(self):
+        """ Inheritance should be applied in the given order """
+        child = self.merge.find('/parent/order/add-first')
+        assert child.data['tag'] == ['one', 'four']
+        child = self.merge.find('/parent/order/remove-first')
+        assert child.data['tag'] == ['one', 'three', 'four']
+
     def test_get(self):
         """ Get attributes """
         assert isinstance(self.wget.get(), dict)
@@ -205,9 +297,9 @@ class TestTree:
 
     def test_show(self):
         """ Show metadata """
-        assert isinstance(self.wget.show(brief=True), type(""))
+        assert isinstance(self.wget.show(brief=True), str)
         assert self.wget.show(brief=True).endswith("\n")
-        assert isinstance(self.wget.show(), type(""))
+        assert isinstance(self.wget.show(), str)
         assert self.wget.show().endswith("\n")
         assert 'tester' in self.wget.show()
 
